/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include "cabin.h"
#include "texture.h"
#include <iostream>
#include "random.h"

void Cabin::log(float radius, float offset, float length) {
	gluQuadricTexture(gluq, GL_TRUE);
	glPushMatrix();
	glRotatef(90.0f, 0.0f, 0.0f, 1.0f);
	glTranslatef(0.0f, 0.0f, offset);
	gluCylinder(gluq, radius, radius, length, gCabinLogSlices, gCabinLogStacks);
	gluDisk(gluq, 0.0f, radius, gCabinLogSlices, 1);
	glTranslatef(0.0f, 0.0f, length);
	glRotatef(180.0f, 0.0f, 1.0f, 0.0f);
	gluDisk(gluq, 0.0f, radius, gCabinLogSlices, 1);
	glPopMatrix();
}

void Cabin::display() {
	if (!mInit) { // What a horrible place to do this initialization
		mCedar.load("data/cedar.raw");
		mStone.load("data/brick.raw");
		mFire[0].load("data/flame1.raw");
		mFire[1].load("data/flame2.raw");
		mFire[2].load("data/flame3.raw");
		mFire[3].load("data/flame4.raw");
		mFirewood.load("data/firewood.raw");
		mInit = true;
	}
	
	glPushMatrix();
	glTranslatef(position().x, position().y, position().z);
	glRotatef(rotation().x, 1.0f, 0.0f, 0.0f);
	glRotatef(rotation().y, 0.0f, 1.0f, 0.0f);
	glRotatef(rotation().z, 0.0f, 0.0f, 1.0f);
	glScalef(0.75f, 0.75f, 0.75f);
	// Peanut butter cabin time.
	mCedar.apply();

	// The wall with the chimney
	glPushMatrix();
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 2.0f, 11.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 4.0f, 7.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 6.0f, 3.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	glPopMatrix();

	// The front wall
	glPushMatrix();
	glTranslatef(20.0f, 0.0f, 0.0f);
	log(1.0f, 0.0f, 5.0f);
	log(1.0f, 10.0f, 5.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 5.0f);
	log(1.0f, 10.0f, 5.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 5.0f);
	log(1.0f, 10.0f, 5.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 5.0f);
	log(1.0f, 10.0f, 5.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 15.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 2.0f, 11.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 4.0f, 7.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 6.0f, 3.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	glPopMatrix();

	// The wall to the right of the door
	glPushMatrix();
	glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
	glTranslatef(-1.0f, -1.0f, -1.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 9.0f);
	log(1.0f, 13.0f, 9.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 9.0f);
	log(1.0f, 13.0f, 9.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);
	glPopMatrix();

	// The wall to the left of the door
	glPushMatrix();
	glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
	glTranslatef(-14.0f, -1.0f, -1.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 9.0f);
	log(1.0f, 13.0f, 9.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 9.0f);
	log(1.0f, 13.0f, 9.0f);	glTranslatef(0.0f, 2.0f, 0.0f);
	log(1.0f, 0.0f, 22.0f);
	glPopMatrix();

	// The roof
	glPushMatrix();
	glTranslatef(10.0f, 14.0f, 11.5f);
	glRotatef(45.0f, 1.0f, 0.0f, 0.0f);
	rectangularPrism(vertex(-12.0f, -1.0f, -6.5f), vertex(12.0f, 1.0f, 6.5f));
	glPopMatrix();
	glPushMatrix();
	glTranslatef(10.0f, 14.0f, 3.5f);
	glRotatef(-45.0f, 1.0f, 0.0f, 0.0f);
	rectangularPrism(vertex(-12.0f, -1.0f, -6.5f), vertex(12.0f, 1.0f, 6.5f));
	glPopMatrix();

	// A floor
	rectangularPrism(vertex(0.0f, 0.0f, 1.0f), vertex(20.0f, 0.5f, 14.0f));

	mStone.apply();

	// The mantle and fireplace area
	rectangularPrism(vertex(0.0f, 0.5f, 4.0f), vertex(4.0f, 1.0f, 11.0f));// bottom
	rectangularPrism(vertex(0.0f, 1.0f, 4.0f), vertex(1.1f, 4.0f, 11.0f));// back
	rectangularPrism(vertex(0.0f, 1.0f, 4.0f), vertex(4.0f, 4.0f, 4.5f));// right
	rectangularPrism(vertex(0.0f, 1.0f, 10.5f), vertex(4.0f, 4.0f, 11.0f));// left
	rectangularPrism(vertex(0.0f, 4.0f, 4.0f), vertex(4.0f, 6.0f, 11.0f));// bottom

	// The chimney
	rectangularPrism(vertex(-3.0f, 0.0f, 4.0f), vertex(0.0f, 8.0f, 11.0f));
	rectangularPrism(vertex(-3.0f, 8.0f, 4.0f), vertex(0.0f, 16.0f, 11.0f));
	rectangularPrism(vertex(-3.0f, 16.0f, 4.0f), vertex(0.0f, 24.0f, 11.0f));

	// Smoke coming out of the smoke stack
	glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	for (int i = 0; i < gCabinSmoke; i++) if (mSmoke[i][4] != 0.0f) {
		glPushMatrix();
		glTranslatef(-1.5f + mSmoke[i][0], 22.0f + mSmoke[i][1], 6.5f + mSmoke[i][2]);
		glColor4f(1.0f, 1.0f, 1.0f, (3.0f - mSmoke[i][3]) / 3.5f);
		glutSolidSphere(mSmoke[i][3], 7, 5);
		glPopMatrix();
	}
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);
	glEnable(GL_LIGHTING);

	
	glPushMatrix();
	glTranslatef(2.0f, 1.0f, 5.5f);
	mFirewood.apply();
	// Some logs on the fire...
	glPushMatrix();
	glTranslatef(0.0f, 0.4f, 0.0f);
	log(0.5f, 0.1f, 4.1f);
	glTranslatef(1.0f, -0.2f, 0.0f);
	glRotatef(4.0f, 0.0f, 1.0f, 0.0f);
	log(0.5f, 0.0f, 4.0f);
	glPopMatrix();
	// A roaring fire
	mFire[mFireCell].apply();
	glDepthMask(GL_FALSE);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glBegin(GL_QUADS);
	// One in the middle 
	glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(1.0f, 0.0f, 4.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(1.0f, 4.0f, 4.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f, 4.0f, 0.0f);
	glEnd();
	mFire[(mFireCell + 1) % 4].apply();
	glBegin(GL_QUADS);
	// One in the front
	glTexCoord2f(1.0f, 1.0f); glVertex3f(1.8f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(1.8f, 0.0f, 4.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(1.8f, 4.0f, 4.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(1.8f, 4.0f, 0.0f);
	// One in the back
	glTexCoord2f(1.0f, 1.0f); glVertex3f(0.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(0.0f, 0.0f, 4.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(0.0f, 4.0f, 4.0f);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(0.0f, 4.0f, 0.0f);
	glEnd();
	glDisable(GL_BLEND);
	glDepthMask(GL_TRUE);
	glPopMatrix();

	// Badger badger badger
	glPopMatrix();
}

void Cabin::rectangularPrism(vertex c1, vertex c2) {
	glBegin(GL_QUADS);
	glNormal3f(1.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c2.x, c1.y, c1.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c2.x, c2.y, c1.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c2.x, c2.y, c2.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c2.x, c1.y, c2.z);
	glNormal3f(-1.0f, 0.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c1.x, c1.y, c1.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c1.x, c1.y, c2.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c1.x, c2.y, c2.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c1.x, c2.y, c1.z);
	glNormal3f(0.0f, 1.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c1.x, c2.y, c1.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c1.x, c2.y, c2.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c2.x, c2.y, c2.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c2.x, c2.y, c1.z);
	glNormal3f(0.0f, -1.0f, 0.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c2.x, c1.y, c1.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c2.x, c1.y, c2.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c1.x, c1.y, c2.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c1.x, c1.y, c1.z);
	glNormal3f(0.0f, 0.0f, 1.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c2.x, c1.y, c2.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c2.x, c2.y, c2.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c1.x, c2.y, c2.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c1.x, c1.y, c2.z);
	glNormal3f(0.0f, 0.0f, -1.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f(c2.x, c1.y, c1.z);
	glTexCoord2f(1.0f, 0.0f); glVertex3f(c1.x, c1.y, c1.z);
	glTexCoord2f(1.0f, 1.0f); glVertex3f(c1.x, c2.y, c1.z);
	glTexCoord2f(0.0f, 1.0f); glVertex3f(c2.x, c2.y, c1.z);
	glEnd();
}

void Cabin::update(float t) {
	static float cumTime = 0.0f;
	int oneFree = -1;
	bool makeNew = true;
	Random r;
	cumTime += t;

	if (cumTime > 0.2f) {
		cumTime = 0.0f;
		mFireCell++;
		if (mFireCell >= 4) mFireCell = 0;
	}

	for (int i = 0; i < gCabinSmoke; i++) { // magic numbers galore
		if (mSmoke[i][3] != 0.0f) {
			mSmoke[i][3] += 0.05f * t;
			mSmoke[i][1] += 0.4f * t + r.floating(0.0f, 0.01f);
			mSmoke[i][0] -= 0.2f * t * mSmoke[i][3];
		}
		if (mSmoke[i][3] > 3.0f || mSmoke[i][3] == 0.0f) oneFree = i;
		if (mSmoke[i][3] != 0.0f & mSmoke[i][3] < 0.2f) makeNew = false;
	}

	if (makeNew && oneFree != -1) {
		mSmoke[oneFree][0] = -1.0f;
		mSmoke[oneFree][1] = 2.0f;
		mSmoke[oneFree][2] = -1.0f;
		mSmoke[oneFree][3] = 0.1f;
	}
}
