/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include <GL/glut.h>
#include "terrain.h"
#include "object.h"
#include "snowfall.h"
#include "texture.h"
#include "treeline.h"
#include "cabin.h"
#include "movingobject.h"
#include "random.h"
#include "sled.h"
#include "trees.h"
#include "clock.h"
#include <list>

#define ABS(a) (((a)>0)?(a):(-(a)))

struct BoxRegion {
	vertex a;
	vertex b;
	bool inside;
};

int gWidth;
int gHeight;
Terrain gTerrain(32, 32, 200.0f, 15.0f, 200.0f);
Object gCamera;
float gCameraJump = 0.0f;
float gCameraHeight = 3.5f;
char gKeys[256];
Snowfall gSnowfall(1024);
Texture gSnow;
Texture gSnowflake;
Texture gTrees;
bool gGrounded = false;
TreeLine gTreeLine;
float gVelocity = 10.0f;
float gTurnSpeed = 90.0f;
Cabin gCabin;
bool gFullScreen;
std::list<MovingObject> gSnowballs;
std::list<BoxRegion> gBounds;
bool gCursorHidden = false;
Sled gSled;
int gViewMode = 0;
Clock gClock;
bool gRButtonDown = false;
float gTreePos[5][3] = {
	{128.0, 0.0, 147.0},
	{118.577, 0.0, 108.391},
	{24.6183, 0.0, 131.983},
	{25.4762, 0.0, 81.313},
	{162.804, 0.0, 14.7754}
};

bool validPosition(vertex p) {
	for (std::list<BoxRegion>::iterator i = gBounds.begin(); i != gBounds.end(); i++)
		if ((i->a.x <= p.x && i->a.y <= p.y && i->a.z <= p.z
			&& i->b.x >= p.x && i->b.y >= p.y && i->b.z >= p.z) != i->inside)
			return false;
	return true;
}

void init() {
	glClearColor(0.2f, 0.2f, 0.25f, 1.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_BLEND);
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	gTerrain.load("data/terrain.raw", 64, 64, 0.0f, 1.0f);
	gSnowfall.spawn(vertex(-50.0f, 50.0f, -50.0f), vertex(250.0f, 100.0f, 250.0f));
	gSnowfall.limit(vertex(-50.0f, 0.0f, -50.0f), vertex(250.0f, 100.0f, 250.0f), true);
	gSnowfall.limit(vertex(100.0f, 0.0f, 128.0f), vertex(122.0f, 24.0f, 142.0f), false);
	gCabin.position() = vertex(100.0f, gTerrain.elevation(100.0f, 128.0f), 128.0f);
	gCamera = gCabin;
	gCamera.translate(vertex(-5.0f, 0.0f, -5.0f));
	gSled.position() = gCamera.position();
	BoxRegion boxAll = {vertex(1.0f, 0.0f, 1.0f), vertex(199.0f, 100.0f, 199.0f), true}; gBounds.push_back(boxAll);
	BoxRegion boxCabinWall1 = {vertex(99.0f, 0.0f, 127.0f), vertex(101.0f, 24.0f, 140.0f), false}; gBounds.push_back(boxCabinWall1);
	BoxRegion boxCabinWall2 = {vertex(99.0f, 0.0f, 127.0f), vertex(117.0f, 24.0f, 130.0f), false}; gBounds.push_back(boxCabinWall2);
	BoxRegion boxCabinWall3 = {vertex(99.0f, 0.0f, 137.0f), vertex(117.0f, 24.0f, 140.0f), false}; gBounds.push_back(boxCabinWall3);
	BoxRegion boxCabinWall4 = {vertex(114.0f, 0.0f, 127.0f), vertex(117.0f, 24.0f, 132.0f), false}; gBounds.push_back(boxCabinWall4);
	BoxRegion boxCabinWall5 = {vertex(114.0f, 0.0f, 135.0f), vertex(117.0f, 24.0f, 140.0f), false}; gBounds.push_back(boxCabinWall5);
	BoxRegion boxChimney = {vertex(97.0f, 0.0f, 130.0f), vertex(104.0f, 24.0f, 137.0f), false}; gBounds.push_back(boxChimney);

	for (int i = 0; i < 5; i++) {
		gTreePos[i][1] = gTerrain.elevation(gTreePos[i][0], gTreePos[i][2]) - 1.0f;
		BoxRegion tree = {
			vertex(gTreePos[i][0] - 2.5f, 0.0f, gTreePos[i][2] - 2.5f),
			vertex(gTreePos[i][0] + 2.5f, 20.0f,	gTreePos[i][2] + 2.5f),
			false};
		gBounds.push_back(tree);
	}
	gSnow.load("data/snow.raw");
	gSnowflake.load("data/snowflake.raw");
	gTrees.load("data/trees.raw");
	gCursorHidden = true;
	glutSetCursor(GLUT_CURSOR_NONE);
	glutWarpPointer(gWidth / 2, gHeight / 2);
	float grey[4] = {0.75f, 0.75f, 0.75f, 1.0f};
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, grey);
}

void timer(int t) {
	float elapsed = gClock.elapsed();
	gClock.update();
	Random r;
	vertex scale = gTerrain.scale();
	vertex initPos = gCamera.position();
	Object *pObject = &gCamera;
	vertex initSledPos = gSled.position();
	vertex groundNormal = gTerrain.normal(gSled.position().x, gSled.position().z);
	float velocity = gVelocity * elapsed;
	float turnSpeed = gTurnSpeed * elapsed;

	if (gViewMode == 1) pObject = &gSled;
	if (gKeys[int('a')]) pObject->translate(vertex(-velocity, 0.0f, 0.0f));
	if (gKeys[int('d')]) pObject->translate(vertex(velocity, 0.0f, 0.0f));
	if (gKeys[int('s')]) pObject->translate(vertex(0.0f, 0.0f, velocity));
	if (gKeys[int('w')]) pObject->translate(vertex(0.0f, 0.0f, -velocity));
	if (gKeys[int('q')]) pObject->rotate(vertex(0.0f, turnSpeed, 0.0f));
	if (gKeys[int('e')]) pObject->rotate(vertex(0.0f, -turnSpeed, 0.0f));

	if (gKeys[int(' ')] || gRButtonDown) {
		vertex gradient = vertex(groundNormal.x, 0.0f, groundNormal.z);
		gSled.velocity() += gradient;
		gSled.advance(elapsed);
		float corr = dotProduct(gSled.direction(), normalize(gSled.velocity()));
		corr = corr * corr;
		gSled.velocity() += -elapsed * 0.1f * gSled.velocity();
		if (corr < 0.75f)
			gSled.velocity() += -elapsed * 10.0f * corr * gSled.velocity();
	}
	
	if (!validPosition(gCamera.position())) gCamera.position() = initPos;
	if (!validPosition(gSled.position())
		|| !validPosition(gSled.position() + vertex(1.0f, 0.0f, 1.5f))
		|| !validPosition(gSled.position() + vertex(1.0f, 0.0f, -1.5f))
		|| !validPosition(gSled.position() + vertex(-1.0f, 0.0f, -1.5f))
		|| !validPosition(gSled.position() + vertex(1.0f, 0.0f, 0.0f))
		|| !validPosition(gSled.position() + vertex(-1.0f, 0.0f, 0.0f))
		|| !validPosition(gSled.position() + vertex(0.0f, 0.0f, -1.5f))
		|| !validPosition(gSled.position() + vertex(0.0f, 0.0f, 1.5f))
		|| !validPosition(gSled.position() + vertex(-1.0f, 0.0f, 1.5f))) {
		gSled.position() = initSledPos;
		gSled.velocity() = vertex(0.0f, 0.0f, 0.0f);
	}

	float terrainHeight = gTerrain.elevation(gCamera.position().x,	gCamera.position().z) + gCameraHeight;
	gCamera.position().y += gCameraJump * elapsed;
	gCameraJump -= 20.0f * elapsed;
	if (gCamera.position().y < terrainHeight) {
		gCamera.position().y = terrainHeight;
		gGrounded = true;
	}

	std::list<MovingObject>::iterator i = gSnowballs.begin();
	while (i != gSnowballs.end()) {
		i->advance(elapsed);
		i->velocity() += elapsed * vertex(0.0f, -20.0f, 0.0f);
		if (!validPosition(i->position()) || (i->position().y < gTerrain.maxHeight(int(i->position().x),
			int(i->position().z))))
			i = gSnowballs.erase(i);
		else
			i++;
	}

	gCabin.update(0.1f);
	gSnowfall.update(elapsed * vertex(-0.1f, -5.0f, 0.0f));

	float heights[4] = {
		gTerrain.elevation(gSled.position().x + 0.5f, gSled.position().z + 1.5f),
		gTerrain.elevation(gSled.position().x + 0.5f, gSled.position().z - 1.5f),
		gTerrain.elevation(gSled.position().x - 0.5f, gSled.position().z - 1.5f),
		gTerrain.elevation(gSled.position().x - 0.5f, gSled.position().z + 1.5f),
	};
	float safe = 0.0f; //(heights[0] + heights[1] + heights[2] + heights[3]) / 4;
	for (int i = 0; i < 4; i++) if (heights[i] > safe) safe = heights[i];
	gSled.position().y = safe;
	gSled.tiltVector() = crossProduct(groundNormal, vertex(0.0f, 1.0f, 0.0f));
	gSled.tilt() = -180.0f / 3.14159f * acosf(dotProduct(groundNormal, vertex(0.0f, 1.0f, 0.0f)));

	glutPostRedisplay();
	glutTimerFunc(10, timer, 0);
}

void display() {
	if (gViewMode == 1) {
		gCamera.position() = gSled.position() + vertex(0.0f, 1.0f, 0.0f);
		gCamera.rotation() = gSled.rotation();
	}
	float pos[4] = {20.0f, 30.0f, 20.0f, 0.0f};
	vertex rot = gCamera.rotation();
	vertex tr = gCamera.position();
	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);
	glPushMatrix();
	glRotatef(-rot.x, 1.0f, 0.0f, 0.0f);
	glRotatef(-rot.y, 0.0f, 1.0f, 0.0f);
	glRotatef(-rot.z, 0.0f, 0.0f, 1.0f);
	glTranslatef(-tr.x, -tr.y, -tr.z);

	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	gSnow.apply();
	gTerrain.display();

	gTrees.apply();
	gTreeLine.display(gTerrain);

	Texture::none();
	for (int i = 0; i < 2; i++)
		mapleTree(gTreePos[i][0], gTreePos[i][1], gTreePos[i][2]);
	for (int i = 3; i < 5; i++)
		pineTree(gTreePos[i][0], gTreePos[i][1], gTreePos[i][2]);

	gCabin.display();

	gSled.display();

	gSnow.apply();
	for (std::list<MovingObject>::iterator i = gSnowballs.begin();
		i != gSnowballs.end(); i++) {
		glPushMatrix();
		glTranslatef(i->position().x, i->position().y, i->position().z);
		glutSolidSphere(0.25f, 4, 3);
		glPopMatrix();
	}

	vertex groundNormal = gTerrain.normal(gSled.position().x, gSled.position().z);
	vertex cross = normalize(crossProduct(groundNormal, vertex(0.0f, 1.0f, 0.0f)));

	// The snowflakes must be drawn last
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	gSnowflake.apply();
	gSnowfall.display();
	glDisable(GL_BLEND);

	glutReportErrors();
	
	glutSwapBuffers();
	glPopMatrix();
}

void reshape(int w, int h) {
	glViewport(0, 0, gWidth = w, gHeight = h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0f, float(w)/float(h?h:1.0f), 0.5f, 300.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void jump() {
	if (gGrounded) {
		gCameraJump = 5.0f;
		gGrounded = false;
	}
}

void shoot() {
	MovingObject snowball;
	snowball.position() = gCamera.position();
	snowball.accelerate(30.0f, gCamera.rotation());
	gSnowballs.push_back(snowball);
}

void keyboard(unsigned char key, int x, int y) {
	key = tolower(key);
	gKeys[key] = true;

	switch (key) {
	case 27: // Escape
		glutSetCursor(GLUT_CURSOR_INHERIT);
		gCursorHidden = false;
		break;
	case ' ':
		jump();
		gSled.velocity() = vertex(0.0f, 0.0f, 0.0f);
		break;
	case 'z': shoot(); break;
	case 'p':
		std::cout << "(" << gCamera.position().x << ", " << gCamera.position().y
			<< ", " << gCamera.position().z << ")\n";
		break;
	case 'f':
		if ((gFullScreen = !gFullScreen)) {
			glutFullScreen();
		} else {
			glutReshapeWindow(640, 480);
		}
		break;
	case 'o':
		if (gViewMode == 0) gViewMode = 1;
		else if (gViewMode == 1) gViewMode = 0;
		break;
	case 'x':
		exit(0);
		break;
	}
}

void keyboardUp(unsigned char key, int x, int y) {
	gKeys[key] = false;
}

void motion(int x, int y) {
	if (!gCursorHidden) return;
	gCamera.rotation().y += 180.0f * (gWidth / 2 - x) / gWidth;
	gCamera.rotation().x += 180.0f * (gHeight / 2 - y) / gHeight;
	if (gViewMode == 1) gSled.rotation().y += 180.0f * (gWidth / 2 - x) / gWidth;
	if (gCamera.rotation().x > 90.0f) gCamera.rotation().x = 90.0f;
	if (gCamera.rotation().x < -90.0f) gCamera.rotation().x = -90.0f;
	if (x != gWidth / 2 || y != gHeight / 2)
		glutWarpPointer(gWidth / 2, gHeight / 2);
}

void mouse(int button, int state, int x, int y) {
	if (button == GLUT_RIGHT_BUTTON) gRButtonDown = state == GLUT_DOWN;
	if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) jump();
	if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) {
		if (gViewMode == 0) gViewMode = 1;
		else if (gViewMode == 1) gViewMode = 0;
	}
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		if (!gCursorHidden) {
			glutSetCursor(GLUT_CURSOR_NONE);
			gCursorHidden = true;
		} else {
			shoot();
		}
	}
}

int main(int argc, char *argv[]) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA|GLUT_DOUBLE|GLUT_DEPTH);
	glutInitWindowSize(gWidth = 640, gHeight = 480);
	glutCreateWindow(argv[0]);
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutPassiveMotionFunc(motion);
	glutMotionFunc(motion);
	glutMouseFunc(mouse);
	glutTimerFunc(10, timer, 0);
	glutIgnoreKeyRepeat(true);
	init();
	glutMainLoop();
	return 0;
}
