/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include "snowfall.h"
#include "random.h"
#include <GL/gl.h>

Snowfall::Snowfall(int particles) {
	mParticles.resize(particles);
	for (size_t i = 0; i < mParticles.size(); i++)
		mParticles[i].alive = false;
}

void Snowfall::update(vertex motion) {
	Random r;

	// Move the funky particles
	for (size_t i = 0; i < mParticles.size(); i++)
		if (mParticles[i].alive) {
			mParticles[i].p += motion;
			if ((mParticles[i].rot += 1.0f) > 360.0f) mParticles[i].rot = 0.0f;
		}
	
	// Check constraints and kill flakes
	for (size_t i = 0; i < mParticles.size(); i++) {
		if (mParticles[i].alive) {
			for (size_t j = 0; j < mConstraints.size(); j++) {
				if (inside(mParticles[i].p, mConstraints[j].p1, mConstraints[j].p2) != mConstraints[j].inside)
					mParticles[i].alive = false;
			}
		}
	}

	// Bring back the dead flakes
	for (size_t i = 0; i < mParticles.size(); i++) {
		if (!mParticles[i].alive) {
			mParticles[i].p.x = r.floating(mSpawn[0].x, mSpawn[1].x);
			mParticles[i].p.y = r.floating(mSpawn[0].y, mSpawn[1].y);
			mParticles[i].p.z = r.floating(mSpawn[0].z, mSpawn[1].z);
			mParticles[i].r = normalize(vertex(r.floating(-1.0f, 1.0f),
				r.floating(-1.0f, 1.0f), r.floating(-1.0f, 1.0f)));
			mParticles[i].rot = r.floating(0.0f, 360.0f);
			mParticles[i].alive = true;
		}
	}
}

void Snowfall::spawn(vertex p1, vertex p2) {
	mSpawn[0] = p1;
	mSpawn[1] = p2;
}

void Snowfall::limit(vertex p1, vertex p2, bool inside) {
	Box b = {p1, p2, inside};
	mConstraints.push_back(b);
}

void Snowfall::display() {
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	glDisable(GL_LIGHTING);
	for (size_t i = 0; i < mParticles.size(); i++) {
		if (mParticles[i].alive) {
			glPushMatrix();
			glTranslatef(mParticles[i].p.x, mParticles[i].p.y, mParticles[i].p.z);
			glRotatef(mParticles[i].rot,
				mParticles[i].r.x, mParticles[i].r.y, mParticles[i].r.z);
			glBegin(GL_TRIANGLES);
			glTexCoord2f(0.0f, 0.0f); glVertex3f(0.0f, 0.0f, 0.0f);
			glTexCoord2f(2.0f, 0.0f); glVertex3f(1.0f, 0.0f, 0.0f);
			glTexCoord2f(0.0f, 2.0f); glVertex3f(0.0f, 1.0f, 0.0f);
			glEnd();
			glPopMatrix();
		}
	}
	glEnable(GL_LIGHTING);
}
