/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#ifndef TERRAINH
#define TERRAINH

#include "matrix.h"
#include "vertex.h"

struct TerrainValue
{
	float height;
	vertex normal;
};

class Terrain : public Matrix<TerrainValue>
{
public:
struct TextureInfo
	{
		float height;
		unsigned int textures[5]; // = {solid, one quarter, one half, three quarters, 1/2 diagonal}
	};
private:
	vertex terrainScale;
	vertex surfaceNormal(unsigned int x, unsigned int y, bool one);
public:
	Terrain(unsigned int width = 64, unsigned int height = 64, float xScale = 1.0f, float yScale = 1.0f, float zScale = 1.0f);
	void clear(float zero);
	void generate(int iterations, float floor, float ceiling);
	void load(const char *file, int width, int height, float floor, float ceiling);
	void generateNormals(unsigned int left, unsigned int bottom, unsigned int right, unsigned int top);
	void generateTexture(unsigned int left, unsigned int bottom, unsigned int right, unsigned int top);
	void scale(float x, float y, float z);
	vertex scale();
	
	// Additions to improve collision detection with the ground.
	float maxHeight(unsigned int x, unsigned int y);
	float minHeight(unsigned int x, unsigned int y);
	float elevation(float x, float y);
	vertex normal(float x, float y);

	void sphereDeform(const vertex &position, float radius);
	void display();
};

#endif
