/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include "trees.h"
#include <GL/glut.h>

float leaves[4] = {0.0f, 0.5f, 0.0f, 0.0f};
float bark[4] = {0.4f, 0.2f, 0.0f, 0.0f};
float black[4] = {0.0f, 0.0f, 0.0f, 0.0f};
float white[4] = {1.0f, 1.0f, 1.0f, 1.0f};

// A pine tree
void pineTree(float x, float y, float z) {
	GLUquadric *gluq = gluNewQuadric();
	glPushAttrib(GL_LIGHTING_BIT);

	const int parts = 5;

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, bark);
	glPushMatrix();
	glTranslatef(x, y, z);
	glScalef(4.0f, 4.0f, 4.0f);
	glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
	gluCylinder(gluq, 0.1f, 0.1f, 0.5f, parts, 1); // trunk
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, leaves);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.5f, 0.3f, 0.5f, parts, 1); // tier 1
	gluDisk(gluq, 0.0f, 0.5f, parts, 1);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.4f, 0.2f, 0.5f, parts, 1); // tier 2
	gluDisk(gluq, 0.0f, 0.4f, parts, 1);
	glTranslatef(0.0f, 0.0f, 0.5f);
	
	gluCylinder(gluq, 0.3f, 0.0f, 0.5f, parts, 1); // tier 3
	gluDisk(gluq, 0.0f, 0.3f, parts, 1);
	glPopMatrix();

	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);

	glPopAttrib();
	gluDeleteQuadric(gluq);
}

// The trees are fractal trees, as best as I felt like managing on the fly
void treeFractal(GLUquadric *gluq, int levels, float radius, float height) {
	const float scaleDown = 0.9f;
	const int branches = 3;
	if (levels < 0) { // end condition
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, leaves);
		gluSphere(gluq, height*0.85f, 5, 5);
		return;
	}
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, bark);
	gluCylinder(gluq, radius, scaleDown * radius, height, 5, 1);
	for (int i = 0; i < branches; i++) {
		glPushMatrix();
		glTranslatef(0.0f, 0.0f, height);
		glRotatef(i * 360.0f / branches, 0.0f, 0.0f, 1.0f);
		glRotatef(30.0f, 0.0f, 1.0f, 0.0f);
		treeFractal(gluq, levels - 1, radius * scaleDown, height * scaleDown); // recurse
		glPopMatrix();
		if (levels - 1 < 0) return;
	}
}

// Draw a maple tree
void mapleTree(float x, float y, float z) {
	GLUquadric *gluq = gluNewQuadric();
	glPushAttrib(GL_LIGHTING_BIT);
	glPushMatrix();
	glTranslatef(x, y, z);
	glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
	treeFractal(gluq, 2, 1.0f, 3.0f);
	glPopMatrix();
	glPopAttrib();
	gluDeleteQuadric(gluq);
}
