/* Cory McWilliams   *
 * mcwilc@rpi.edu    *
 * December 12, 2003 */

#include "vertex.h"
#include <cmath>

float abs(const vertex &a)
{
	return static_cast<float>(sqrt(a.x * a.x + a.y * a.y + a.z * a.z));
}

vertex normalize(const vertex &a)
{
	float r = abs(a);
	vertex v(a.x / r, a.y / r, a.z / r);
	return v;
}

vertex crossProduct(const vertex &a, const vertex &b)
{
	vertex v(a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x);
	return v;
}

float dotProduct(const vertex &a, const vertex &b)
{
	return a.x * b.x + a.y * b.y + a.z * b.z;
}

vertex3 operator+(vertex3 const &a, vertex3 const &b)
{
	return vertex(a.x + b.x, a.y + b.y, a.z + b.z);
}

vertex3 operator-(vertex3 const &a, vertex3 const &b)
{
	return vertex(a.x - b.x, a.y - b.y, a.z - b.z);
}

vertex3 &operator+=(vertex3 &a, vertex3 const &b)
{
	a = a + b;
	return a;
}

vertex3 operator/(vertex3 const &a, float b)
{
	vertex3 v(a.x / b, a.y / b, a.z / b);
	return v;
}

vertex3 &operator/=(vertex3 &a, float b)
{
	return a = a / b;
}

vertex3 Vertex3(float x, float y, float z)
{
	vertex3 v(x, y, z);
	return v;
}

vertex3 operator*(float a, const vertex3 &r)
{
	vertex3 v(a * r.x, a * r.y, a * r.z);
	return v;
}

vertex3 operator*(const vertex3 &r, float a)
{
	vertex3 v(a * r.x, a * r.y, a * r.z);
	return v;
}

float relativeComponent(const vertex &r, const vertex &x, float absx)
{
	return dotProduct(r, x) / absx;
}

float relativeVelocity(const vertex &PA, const vertex &PB, const vertex &VA, const vertex &VB)
{
	float absx = abs(PB - PA);
	float va = relativeComponent(VA, PA - PB, absx);
	float vb = relativeComponent(VB, PB - PA, absx);
	return vb + va;
}

bool inside(vertex p, vertex e1, vertex e2) {
	return p.x >= e1.x && p.x <= e2.x
		&& p.y >= e1.y && p.y <= e2.y
		&& p.z >= e1.z && p.z <= e2.z;
}
